#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
userprovided library
~~~~~~~~~~~~~~~~~~~~~
Source: https://github.com/RuedigerVoigt/userprovided
(c) 2020-2021 Rüdiger Voigt
Released under the Apache License 2.0
"""

from userprovided import _version
from userprovided import date
from userprovided import hash  # Deprecated / logs warning until v 1.0
from userprovided import hashing
from userprovided import mail
from userprovided import parameters
from userprovided import url


NAME = "userprovided"
__version__ = _version.__version__
__author__ = "Rüdiger Voigt"
